#Written by: Brennan Marsh-Armstrong
#Name: Local Min-Max Normalized Image Processing and Vasculature Profiling
#Last Udated: 8.11.19
import scipy.misc as sp
import numpy as np
from matplotlib import pyplot as plt
import json,cv2
from scipy.signal import find_peaks
from scipy import ndimage
from skimage.morphology import medial_axis, skeletonize, skeletonize_3d
from scipy.spatial import Delaunay

#=======================================================================================================================
#\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
#=======================================================================================================================

class filehandler():
    """Class containing methods for reading files"""

    #===================================================================================================================
    #===================================================================================================================

    def readImage(self,picname,yRange,xRange):
        """
        Purpose:
            - Read image file, crops to desired area, and returns numpy array of image
        Parameters:
            -picname: image file name including extension (str)
            -yRange: tuple [min,max] y range of image to be cropped (tuple)
            -xRange: tuple [min,max] x range of image to be cropped (tuple)
        Returns:
            -Numpy array representation of image
        """
        pic=sp.imread(picname,flatten=True)
        if yRange is None: yRange=(0,int(pic.shape[0]))
        if xRange is None: xRange=(0,int(pic.shape[1]))
        return np.array(pic[yRange[0]:yRange[1],xRange[0]:xRange[1]])

    #===================================================================================================================
    #===================================================================================================================

    def readConfig(self,fName='config_LMNM.txt'):
        """
        Purpose:
            - Reads the config text file and returns a dictionary of parameter:value pairs
        Parameters:
            -fName: file name of config file (str)
        Returns:
            -dictionary of parameterName:value pairs
        """
        with open(fName) as f: return json.load(f)

#=======================================================================================================================
#\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
#=======================================================================================================================

class imageProcessor():
    """Class containing collection of image processing methods"""

    #===================================================================================================================
    #===================================================================================================================

    def floorRaise(self,pic,exclusionMin):
        """
        Purpose:
            - Reduce low-intensity noise by raising all pixel values to a minimum value.
        Parameters:
            -pic: numpy array of the image being processed (np arr)
            -exclusionMin: minimum value all pixels are increased to (int)
        Returns:
            -processed numpy array with min pixel value increased
        """
        if exclusionMin is not None:
            comparArr=np.full(pic.shape,exclusionMin)
            return np.array([val-exclusionMin for val in np.maximum(comparArr,pic)])
        else: return pic

    #===================================================================================================================
    #===================================================================================================================

    def loopGaussBlurring(self,pic,iterations,size):
        """
        Purpose:
            - Repeatedly gaussian blurs an image (sigma=kernel size)
        Parameters:
            -pic: numpy array of the image being processed (np arr)
            -iterations: number of times the blur occurs (int)
            -size: kernel of the blurring (odd int)
        Returns:
            -processed blurred numpy array
        """
        if iterations is not None:
            for iter in range(0,iterations):
                pic=cv2.GaussianBlur(pic,(size,size),0)
        return pic

    #===================================================================================================================
    #===================================================================================================================

    def backgroundReduction(self,pic,blurIt,blurS,frMin):
        """
        Purpose:
            - Calls loopGaussBlurring and floorRaise together. See those methods for clarification
        Parameters:
            -pic: numpy array of the image being processed (np arr)
            -blurIt: number of times the blur occurs (int)
            -blurS: kernel of the blurring (odd int)
            -frMin: minimum value all pixels are increased to (int)
        Returns:
            -processed numpy array image with background reduction
        """
        picBR = self.loopGaussBlurring(pic,blurIt,blurS)
        picBR = self.floorRaise(picBR,frMin)
        return picBR

    #===================================================================================================================
    #===================================================================================================================

    def processHorizontal1(self,im,highCutoffR,lowCutoffR):
        """
        Purpose:
            - Local Min Max Normalization Masking type1 (LMNM), issolates vertical vasculature in image. Values at or above highCutoffR are 255, at or below lowCutoffR are 0, and those between are scaled.
        Parameters:
            -im: numpy array of the image being processed (np arr)
            -highCutoffR: ratio of max-min difference scaled to 255 (float)
            -lowCutoffR: ratio of max-min difference scaled to 0 (float)
        Returns:
            -numpy array image with vertical vasculature emphasized
        """
        processed=np.zeros(im.shape)

        #Loops through all horizontal pixel slices in image
        for j in range(0,im.shape[0]-1):
            #Identifies local maximums and minimums
            working=np.array(im[j:j+1,:][0])
            workingfiltered=np.zeros(working.shape)
            maxs=find_peaks(working)[0].tolist()
            mins=find_peaks(-working)[0].tolist()

            #Combines maximums and minimums in alternating order
            maxmins=[0]
            while (len(maxs)+len(mins))>1:
                if maxs[0]<mins[0]: maxmins.append(maxs.pop(0))
                else: maxmins.append(mins.pop(0))
            maxmins=maxmins+maxs+mins

            #Sorts through sequential pairs in maxmins (minimum maximum pairs)
            for p1,p2 in zip(maxmins[:-1],maxmins[1:]):
                #Figures out which is a maximum and which is a minimum
                min=working[p1]
                max=working[p2]
                if min>max:
                    max = working[p1]
                    min = working[p2]
                highcutoff=(max-min)*highCutoffR+min
                lowcutoff=(max-min)*lowCutoffR+min
                #Scales pixel-slice between the two cutoffs
                for i in range(p1,p2):
                    if working[i]>=highcutoff: workingfiltered[i]=255
                    elif working[i]<=lowcutoff: workingfiltered[i]=0
                    else: workingfiltered[i]=(working[i]-lowcutoff)/(highcutoff-lowcutoff)*255

            #Displays how the linear slice has been processed. Turn displayOn to True to activate
            displayOn = False
            if displayOn:
                a=plt.subplot2grid((1, 2), (0, 0))
                plt.plot(working)
                plt.plot(maxmins,[working[i] for i in maxmins],'ro')
                plt.subplot2grid((1, 2), (0, 1),sharex=a)
                plt.plot(workingfiltered)
                plt.plot(maxmins,[workingfiltered[i] for i in maxmins],'bo')
                plt.show()

            #Puts processed slice back into image
            processed[j,:]=workingfiltered

        return processed

    #===================================================================================================================
    #===================================================================================================================

    def processHorizontal2(self,im,peakSpacingMin,peakHeighMin,ratioCutoff):

        """
        Purpose:
            - Local Min Max Normalization Masking type1 (LMNM), issolates vertical vasculature in image. Values above ratioCutoff are masked those below are not.
        Parameters:
            -im: numpy array of the image being processed (np arr)
            -peakSpacingMin: Minimum horizontal pixel distance between two peaks that is allowed (int)
            -peakHeighMin: Minimum pixel value difference between Min and Max that is allowed (int)
            -ratioCutoff: ratio of max-min difference that is masking cutoff
        Returns:
            -numpy array image with vertical vasculature emphasized
        """

        processed=np.zeros(im.shape)

        #Loops through all horizontal pixel slices in image
        for j in range(0,im.shape[0]-1):
            #Identifies local maximums and minimums
            working=np.array(im[j:j+1,:][0])
            workingfiltered=np.zeros(working.shape)
            maxs=find_peaks(working)[0].tolist()
            mins=find_peaks(-working)[0].tolist()

            #Combines maximums and minimums in alternating order
            maxmins=[0]
            while (len(maxs)+len(mins))>1:
                if maxs[0]<mins[0]: maxmins.append(maxs.pop(0))
                else: maxmins.append(mins.pop(0))
            maxmins=maxmins+maxs+mins

            #Filters maximum minimum list to remove peaks too close together or too small
            maxminsProcessed=[]
            for p1,p2,p3 in zip(maxmins[:-2],maxmins[1:-1],maxmins[2:]):
                if abs(p1-p3) > peakSpacingMin and (abs(working[p1]-working[p2]) > peakHeighMin or abs(working[p3]-working[p2]) > peakHeighMin):
                    maxminsProcessed.append(p2)
            maxminsProcessed=[maxmins[0]]+maxminsProcessed+[maxmins[-1]]

            #Cleans up the peak removal above by removing instances of adjacent maximums or adjacent minimums without an interceding min or max respectively
            maxminsProcessed2=[]
            for p1,p2,p3 in zip(maxminsProcessed[:-2],maxminsProcessed[1:-1],maxminsProcessed[2:]):
                dir1='up'
                dir2='up'
                if working[p1] > working[p2]: dir1 = 'down'
                if working[p2] > working[p3]: dir2 = 'down'
                if dir1 is not dir2: maxminsProcessed2.append(p2)
            maxminsProcessed2=[maxminsProcessed[0]]+maxminsProcessed2+[maxminsProcessed[-1]]
            maxminsProcessed=maxminsProcessed2

            #Loop through min-max pairs and normalize the pixels between each by the specified ratio between the max and min
            for p1,p2 in zip(maxminsProcessed[:-1],maxminsProcessed[1:]):
                min=working[p1]
                max=working[p2]
                if min>max:
                    max = working[p1]
                    min = working[p2]
                workingfiltered[p1:p2] = [int(a)*255 for a in [working[i]>ratioCutoff*max for i in range(p1,p2)]]
                processed[j,:]=workingfiltered
        return processed

    #===================================================================================================================
    #===================================================================================================================

    def processHorVerCombine(self,im,procType,peakSpacingMin,peakHeighMin,ratioCutoff,lowCutoffR,highCutoffR):
        """
        Purpose:
            - Run LMNM processing both horizontally and vertically on an image
        Parameters:
            -im: numpy array of the image being processed (np arr)
            -procType: int specifying which LMNM processing to perform (int [1 or 2])
            -remaining_variables: see processHorizontal1 and processHorizontal2 for variable descriptions
        Returns:
            -tuple of the horizontally and vertical LMNM processed images
        """

        if procType is 1:
            xfilt=self.processHorizontal1(im,highCutoffR,lowCutoffR)
            yfilt=np.transpose(self.processHorizontal1(np.transpose(im),highCutoffR,lowCutoffR))
        if procType is 2:
            xfilt=self.processHorizontal2(im,peakSpacingMin,peakHeighMin,ratioCutoff)
            yfilt=np.transpose(self.processHorizontal2(np.transpose(im),peakSpacingMin,peakHeighMin,ratioCutoff))
        return xfilt,yfilt

    #===================================================================================================================
    #===================================================================================================================

    def objSizeExclusion(self,imgIn,minobjsize):
        """
        Purpose:
            - Morphologically process image to identify objects and remove objects smaller than a set size
        Parameters:
            -imgIn: numpy array of the image being processed (np arr)
            -minobjsize: minimum pixel^2 area size objects can be and without being removed (int)
        Returns:
            -numpy binary array with small objects removed
        """

        imgIn=np.copy(imgIn)
        Zlabeled,Nlabels = ndimage.measurements.label(imgIn,structure=[[1,1,1],[1,1,1],[1,1,1]])
        label_size = [(Zlabeled == label).sum() for label in range(Nlabels + 1)]
        for label,size in enumerate(label_size):
            if size < minobjsize:
                imgIn[Zlabeled == label] = 0
        return imgIn

    #===================================================================================================================
    #===================================================================================================================

    def xyCombine(self,imX,imY,loopNum,gSize):
        """
        Purpose:
            - Morphologically process image to identify objects and remove objects smaller than a set size.
        Parameters:
            -imX: horizontally LMNM processed image (np array)
            -imY: vertically LMNM processed image (np array)
            -loopNum: number of times the blur occurs (int)
            -gSize: kernel of the blurring (odd int)
        Returns:
            -blurred combination of LMNM processed vertical and horizontal image (np array)
        """

        #X and Y LMNM images blurred to close small gaps when combined
        imX=self.loopGaussBlurring(imX,loopNum,gSize)
        imY=self.loopGaussBlurring(imY,loopNum,gSize)
        combinedXY=np.maximum(imX,imY)
        return combinedXY

    #===================================================================================================================
    #===================================================================================================================

    def mask(self,im,thresh,sizeExclNum):
        """
        Purpose:
            - Mask the blurred LMNM image back into a binary mask and remove small objects
        Parameters:
            -im: combined blurring of X and Y LMNM processed image (np array)
            -thresh: threshold of image masking (int)
            -sizeExclNum: See objSizeExclusion methods for clarification
        Returns:
            -numpy binary array with small objects removed
        """

        imout=(cv2.threshold(im,thresh,255,cv2.THRESH_BINARY))[1]
        if sizeExclNum is not None:
            imout=self.objSizeExclusion(imout,sizeExclNum)
        return imout

    #===================================================================================================================
    #===================================================================================================================

    def skeletonize(self,im,type):
        """
        Purpose:
            - Skeletionize image binary mask
        Parameters:
            -im: binary vessel image (np array)
            -type: specification of what method of skeletionizing is going to be used (str ['2d', '3d', or 'median'])
            -sizeExclNum: See objSizeExclusion methods for clarification
        Returns:
            -numpy binary array with small objects removed
        """

        #All skeletionization methods used are part of the skimage.morphology package
        if type == '2d':
            skeleton = skeletonize(im/255)
        if type == '3d':
            skeleton = skeletonize_3d(im/255) == 255
        if type == 'median':
            skeleton = medial_axis(im/255, return_distance=False)
        return skeleton

#=======================================================================================================================
#\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
#=======================================================================================================================

class skelProcessor():
    """Class containing collection of methods used to process and convert binary skeleton to graph-representation"""

    #===================================================================================================================
    #===================================================================================================================

    def neighborCountMapper(self,im):
        """
        Purpose:
            -Convert binary skeleton to skeleton with values based on number of adjacent skeleton locations
        Parameters:
            -im: binary vessel skeleton (np array)
        Returns:
            -numpy array with neighbor-count coded vessel skeleton
        """

        impadded=np.full((im.shape[0]+2,im.shape[1]+2),False)
        impadded[1:-1,1:-1]=im
        countSkel=np.zeros(im.shape)
        for y,x in np.ndindex(impadded.shape):
            if impadded[y,x]:
                countSkel[y-1,x-1] = np.sum(impadded[y-1:y+2,x-1:x+2])-1
        return countSkel

    #===================================================================================================================
    #===================================================================================================================

    def shortBranchConnector(self,im,fitLen,fitAngle,findLen):
        """
        Purpose:
            -Complete partial branches by fitting loose ends, extrapolating, and looking for overlap
        Parameters:
            -im: neighbor-coded vessel skeleton (np array)
            -fitLen: number of pixels of the skeleton used to estimate vessel path (int)
            -fitAngle: angle offset from estimated vessel path allowed as a valid path (int)
            -findLen: maximum skeleton-gap length repaired (int)
        Returns:
            -neighbor-coded vessel skeleton with small holes repaired
        """

        impadded=np.full((im.shape[0]+findLen*2,im.shape[1]+findLen*2),0)
        impadded[findLen:-findLen,findLen:-findLen]=im

        #Search through every position in the image
        for y,x in np.ndindex(impadded.shape):
            #If location is skeleton end (has only one adjacent skeleton pixel)
            if impadded[y,x] == 1:
                #Find full vessel to the nearest branch point
                visited = [(y,x)]
                yT,xT=y,x
                while impadded[yT,xT] < 3:
                    for yT2,xT2 in np.ndindex((3,3)):
                        yT2,xT2 = yT2+yT-1,xT2+xT-1
                        if impadded[yT2,xT2] != 0 and (yT2,xT2) not in visited:
                            if impadded[yT2,xT2] < 3: visited.append((yT2,xT2))
                            yT,xT = yT2,xT2
                            break
                fitList = visited[:fitLen]
                #Skips vessels that could not be reliably fit
                if len(list(set([f[1] for f in fitList])))<2: continue #deals with straight horizontal lines (poorly fit)
                if len(fitList)<=3: continue  #deals with short lines (poorly fit)

                #Fit end of vessel to a linear regression
                fit = np.poly1d(np.polyfit([coord[1] for coord in fitList],[coord[0] for coord in fitList],1))

                #Establishes linear regressions rotated a specified number of degrees from the existing vessel-fit
                slopeUpDeg = np.rad2deg(np.arctan2(fit.c[0],1))+fitAngle/2
                slopeDownDeg = np.rad2deg(np.arctan2(fit.c[0],1))-fitAngle/2
                slopeUp = np.tan(np.deg2rad(slopeUpDeg))
                slopeDown = np.tan(np.deg2rad(slopeDownDeg))
                if slopeUpDeg > 90 or slopeUpDeg < -90:
                    slopeUp = -slopeUp
                if slopeDownDeg > 90 or slopeDownDeg < -90:
                    slopeDown = -slopeDown
                fitUp = np.poly1d([slopeUp,visited[0][0]-slopeUp*visited[0][1]])
                fitDown = np.poly1d([slopeDown,visited[0][0]-slopeDown*visited[0][1]])

                #Determine which direction the regressions should extend out of the vessel endpoint
                fitUpX = findLen+1
                fitDownX = findLen+1
                if  fitList[-1][1]>fitList[0][1]: fitUpX,fitDownX = -fitUpX,-fitDownX

                #Generates a convex full from the two angle-offset linear extrapolations of the vessel
                hullPoints = [[visited[0][1],fitUp(visited[0][1])],[visited[0][1]+fitUpX,fitUp(visited[0][1]+fitUpX)],[visited[0][1]+fitDownX,fitDown(visited[0][1]+fitDownX)]]
                hull = Delaunay(hullPoints)

                #Finds all skeleton points within the specified distance from the vessel loose-end and excludes those that are previously-closed loose ends and are not within the convex hull
                possiblePoints = [[xC-findLen+visited[0][1],yC-findLen+visited[0][0]] for yC,xC in np.ndindex((findLen*2+1,findLen*2+1)) if impadded[yC-findLen+visited[0][0],xC-findLen+visited[0][1]]]
                possiblePoints = [p for p in possiblePoints if (p[1],p[0]) not in visited]
                possiblePoints = [possiblePoints[i] for i in range(0, len(possiblePoints)) if (hull.find_simplex(possiblePoints)>=0)[i]]

                #Prioritizes other loose-ends or existing branch-points as locations to have the patch-vessel end at
                if 1 in [impadded[p[1],p[0]] for p in possiblePoints] or 3 in [impadded[p[1],p[0]] for p in possiblePoints]:
                    possiblePoints = [p for p in possiblePoints if impadded[p[1],p[0]]!=2]

                #If a valid location to close the vessel loose end is found, a line is calculated between the two and the vessel is added.
                if len(possiblePoints)>0:
                    ppDistances = [np.sqrt((p[1]-visited[0][0])**2+(p[0]-visited[0][1])**2) for p in possiblePoints]
                    closestP = sorted(zip(possiblePoints,ppDistances), key=lambda x: x[1])[0][0]
                    fitAcross = np.poly1d(np.polyfit([visited[0][1],closestP[0]],[visited[0][0],closestP[1]],1))
                    yChange = abs(closestP[1]-visited[0][0])
                    xChange = abs(closestP[0]-visited[0][1])
                    res = np.maximum(yChange,xChange)
                    fillRange = range(0,res+1)
                    if  fitList[-1][1]>fitList[0][1]: fillRange = range (-res,1)
                    patch = [(int(round(fitAcross(r))),int(round(r))) for r in [float(r)*(float(xChange)/float(res))+visited[0][1] for r in fillRange]]
                    for (a,b) in patch: impadded[a,b] = 2

        #Neighbor-count repeated to resolve inconsistencies caused by processing.
        connectedIm = (impadded[findLen:-findLen,findLen:-findLen]!=0)
        return self.neighborCountMapper(connectedIm)

    #===================================================================================================================
    #===================================================================================================================

    def shortBranchPruner(self,im,minLength):
        """
        Purpose:
            -Prune branches which are too small to be anything other than skeletonization errors
            (This is the wrapper method for the core method. Called twice because calling opens up new loose-ends)
        Parameters:
            -im: neighbor-coded vessel skeleton (np array)
            -minLength: minimum lenght of a vessel that is genuine as opposed to skeletonization error
        Returns:
            -neighbor-coded vessel skeleton (np array) with samll branches pruned off
        """
        prune1 = self.shortBranchPrunerCore(im,minLength)
        return self.shortBranchPrunerCore(prune1,minLength)

    #===================================================================================================================
    #===================================================================================================================

    def shortBranchPrunerCore(self,im,minLength):
        """
        Purpose:
            -Prune branches which are too small to be anything other than skeletonization errors
            (This is the core method)
        Parameters:
            -im: neighbor-coded vessel skeleton (np array)
            -minLength: minimum lenght of a vessel that is genuine as opposed to skeletonization error
        Returns:
            -neighbor-coded vessel skeleton (np array) with samll branches pruned off
        """
        impadded=np.full((im.shape[0]+2,im.shape[1]+2),0)
        impadded[1:-1,1:-1]=im

        #Search through all locations in the image
        for y,x in np.ndindex(impadded.shape):

            special2 = False
            #Determins if it is a special location where 2-neighbor locations are adjacent to 2 branch-points (3+ neighbors).
            #   These are errors that need to be removed
            if impadded[y,x] == 2 and np.count_nonzero(impadded[y-1:y+2,x-1:x+2] == 3) == 2:
                where2 = np.where(impadded[y-1:y+2,x-1:x+2] ==3)
                special2 = np.sqrt((where2[0][0]-where2[0][1])**2+(where2[1][0]-where2[1][1])**2)==1
            #Determins if it is a loose-end vessel
            if impadded[y,x] == 1 or special2:
                visited = [(y,x)]
                yT,xT=y,x
                #Move along loose-end vessel starting at end and finds whole vessel, stopping at branch point
                while impadded[yT,xT] < 3:
                    for yT2,xT2 in np.ndindex((3,3)):
                        yT2,xT2 = yT2+yT-1,xT2+xT-1
                        if impadded[yT2,xT2] != 0 and (yT2,xT2) not in visited:
                            if impadded[yT2,xT2] < 3: visited.append((yT2,xT2))
                            yT,xT = yT2,xT2
                            break
                #If vessel length is too small, it is removed.
                if len(visited) < minLength:
                    for v in visited:
                        impadded[v[0],v[1]]=0

        #Neighbor-count repeated to resolve inconsistencies caused by processing.
        trimmedIm = (impadded[1:-1,1:-1]!=0)
        return self.neighborCountMapper(trimmedIm)

    #===================================================================================================================
    #===================================================================================================================

    def makeSkeletonGraph(self,im):
        """
        Purpose:
            -Convert mask of skeleton to graphical representation of points and branches
        Parameters:
            -im: neighbor-coded vessel skeleton (np array)
        Returns:
            -List of points where each point has reference to all neighboring points on the skeleton
        """

        #Generate list of skeleton points
        impadded=np.full((im.shape[0]+2,im.shape[1]+2),0)
        impadded[1:-1,1:-1]=im
        skelCoords = [(y,x) for y,x in np.ndindex(impadded.shape) if bool(impadded[y,x])]
        skelGraph={}

        #At each skeleton point make a list of all neighboring point addresses and create a Point with the neighbors as that list
        #See Point Class (below) for more information
        for coord in skelCoords:
            y,x = coord
            imAround=impadded[y-1:y+2,x-1:x+2]
            coordOut = (y-1,x-1)
            neighbors = [(int(y+yI-2),int(x+xI-2)) for yI,xI in np.ndindex(imAround.shape) if bool(imAround[yI,xI]) and (yI,xI)!=(1,1)]
            pt = Point(coordOut)
            pt.neighbors = neighbors
            skelGraph[coordOut] = pt
        return skelGraph

    #===================================================================================================================
    #===================================================================================================================

    def makeNodeGraph(self,im,ptGraph):

        """
        Purpose:
            -Process skeleton graph to issolate point types (branch-point, part-of-vessel, or loose-end), as well as
            create list of full branches (each is a lists of points)
        Parameters:
            -im: neighbor-coded vessel skeleton (np array)
            -ptGraph: graphical representation of skeleton
        Returns:
            -tuple with dictionaries with 1) (Y,X) tuple (y,x location of branch point) as key for list of adjacent nodes, and
            2) ((Y1,X1),(Y2,X2)) tuple of tuples (y-x locations of connected branch-points) as key for list of points in the branch
        """

        #Processes skeleton image to mask only branch-points (in the skeleton branch-points can be multiple adjacent points)
        nodeIm = np.bitwise_and(im!=2, im!=0) * 1
        #Morphologically labels each clump of "branch-points" with a different number
        lbls = ndimage.measurements.label(nodeIm,structure=[[1,1,1],[1,1,1],[1,1,1]])
        #Identify the center positions of each branch-point.
        nodePts = self.findNodes(nodeIm,lbls)
        #Find all locations where branches extend from each branch-point clump and assign them to their respective nodePts
        nodeBPts = self.findNodeBranchePts(im,lbls,nodePts)
        #Linking branch points to their connected branch points, reate dictionaries organizing points into branches,
        #   linking branches to their 2 adjacent branch-points
        nodeDataGraph,branchDataGraph = self.createBranches(ptGraph,nodePts,nodeBPts,lbls[0])
        return nodeDataGraph,branchDataGraph

    #===================================================================================================================
    #===================================================================================================================

    def findNodes(self,im,lbls):
        """
        Purpose:
            -Define center-points for multi-pixel vessel branch points.
        Parameters:
            -im: mask of branch-point clumps (np array)
            -lbls: labeled mask of branch-point clumps (np array)
        Returns:
            -list of (y,x) tuples of branch points
        """

        comPts = [(int(round(y)),int(round(x))) for y,x in ndimage.measurements.center_of_mass(im,lbls[0],range(1,lbls[1]+1))]
        return comPts

    #===================================================================================================================
    #===================================================================================================================

    def findNodeBranchePts(self,im,lbls,nPts):
        """
        Purpose:
            -Generates list of starts starts to branches that are connected to each branch-point.
        Parameters:
            -im: neighbor-coded vessel skeleton (np array)
            -lbls: labeled mask of branch-point clumps (np array)
            -nPts: list of (y,x) coordinates of branch points (list[tuples])
        Returns:
            -List in which index is associated label of branch-point # and contents is a list of (y1,x1,y2,x2) tuples in which
            y1 and x1 define where in the branch-point clump a branch comes from and y1,x1 define the first pixel of the branch.
        """

        #creates a list of lists as long as the number of labeled branch-point clumps
        nodeBpLst=[[]] * lbls[1]
        #Search through every index in the labeled mask of branch-point clums and finds all masked pixels
        for y,x in np.ndindex(im.shape):
            if lbls[0][y,x]!=0:
                searchGrid = im[y-1:y+2,x-1:x+2]
                for y2,x2 in np.ndindex(searchGrid.shape):
                    #If there is a branch pixel adjacent to a branch-point clump pixel then it is recorded in a list organized
                    #   by the branch-point it is associated with
                    if searchGrid[y2,x2]==2:
                        nodeBpLst[lbls[0][y,x]-1]=nodeBpLst[lbls[0][y,x]-1]+[(y,x,y2+y-1,x2+x-1)]
        return nodeBpLst

    #===================================================================================================================
    #===================================================================================================================

    def createBranches(self,ptGraph,nodeP,nodeBP,lbl):
        """
        Purpose:
            -The final step in generating a graph-data-type representation of branches and branch-points in the image.
        Parameters:
            -ptGraph: list of Points, which each contain knowledge of all their neighboring points (list)
            -nodeP: list of coordinates for branch-point-clump centers (list)
            -nodeBP: List in which index is associated label of branch-point # and contents is a list of (y1,x1,y2,x2) tuples in which
            y1 and x1 define where in the branch-point clump a branch comes from and y1,x1 define the first pixel of the branch. (list)
            -lbl: labeled mask of branch-point clumps (np array)
        Returns:
            -tuple with dictionaries with 1) (Y,X) tuple (y,x location of branch point) as key for list of adjacent nodes, and
            2) ((Y1,X1),(Y2,X2)) tuple of tuples (y-x locations of connected branch-points) as key for list of points in the branch
        """

        branchGraph = {}
        nodeGraph = dict.fromkeys(nodeP,[])

        # each branch point is assigned a number...
        for i in range(0,len(nodeP)):
            #for each of the first branch pixels attached to the branch point
            for branchStart in nodeBP[i]:

                #Searches down the branch and searches until it finds the branch-point pixel
                startNode = nodeP[i]
                branchExplored = [(branchStart[0],branchStart[1]),(branchStart[2],branchStart[3])]
                currSpotCoord = (branchStart[2],branchStart[3])
                currSpotPoint = ptGraph[currSpotCoord]

                while len(currSpotPoint.neighbors)==2:
                    if currSpotPoint.neighbors[0] in branchExplored and currSpotPoint.neighbors[1] in branchExplored:
                        branchExplored.append(branchExplored[0])
                        break
                    currSpotCoord = [sc for sc in currSpotPoint.neighbors if sc not in branchExplored][0]
                    currSpotPoint = ptGraph[currSpotCoord]
                    branchExplored.append(currSpotCoord)

                #Removes the branch-point pixels at the start of the branch path used to prevent tracing backwards
                snipTail = (branchExplored[-1][0],branchExplored[-1][1],branchExplored[-2][0],branchExplored[-2][1])
                if snipTail in nodeBP[lbl[branchExplored[-1][0],branchExplored[-1][1]]-1]:
                    nodeBP[lbl[branchExplored[-1][0],branchExplored[-1][1]]-1].remove(snipTail)

                #startNode, endNode, and branchExplored are used to update graphs that organize branch-points' connections
                #   to one another and branches joining pairs of branch-points
                endNode = nodeP[lbl[branchExplored[-1][0],branchExplored[-1][1]]-1]
                branchGraph[(startNode,endNode)] = branchExplored
                nodeGraph[startNode] = nodeGraph[startNode]+[endNode]
                nodeGraph[endNode] = nodeGraph[endNode]+[startNode]

        return nodeGraph,branchGraph

#=======================================================================================================================
#\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
#=======================================================================================================================

class skelAnalysis:
    """Class containing collection of image analysis methods for extracting anatomical metrics"""
    #===================================================================================================================
    #===================================================================================================================

    def __init__(self,skelNum,vessels,flowVoids,skelG,nodeG,branchG,dataInclusionMask,p2SRatio):
        """
        Purpose:
            -Constructor method for passing all necessary parameters in for anatomical analysis.
        Parameters:
            -skelNum: neighbor-coded vessel skeleton (np array)
            -vessels: binary vessel mask (np array)
            -flowVoids: binary flowVoid mask (np array) --can just be the inverse of 'vessels' parameter if same threshold used
            -skelG: graph with (y,x) coordinate tuples as keys and points at those locations as values (dict)
            -nodeG: graph with (y,x) coordinate tuples as keys and a list of connected point coordinates as values (dict)
            -branchG: graph with ((y1,x1),(y2,x2)) coordinate pairs as keys and the branch connecting those two points as a value (dict)
            dataInclusionMask: binary mask of area of image to be processed (np array)
            -p2SRatio: ratio conversion for pixel size to real units (float)
        Returns:
            -None
        """
        #Not all parameters passed in are necessary for currently implemented data analysis. Parameters are passed in
        #   so that future data analysis will be easier

        #Notice that all methods in this class are extremely simple. That is by design. Once you have the image in the formats
        #   required for the constructor, extracting anatomical metrics is relatively simple

        #Uses the data exclusion map to remove data that shouldn't be processed.
        self.numberedSkel = skelNum * dataInclusionMask
        for key,value in nodeG.items():
            if dataInclusionMask[key[0],key[1]]==0: del nodeG[key]
        self.nodeGraph = nodeG
        for key,value in branchG.items():
            if dataInclusionMask[key[0][0],key[0][1]]==0 or dataInclusionMask[key[1][0],key[1][1]]==0: del branchG[key]

        #Assigns parameters to private variables so they can be accessed by methods
        self.branchGraph = branchG
        self.vessels = vessels * dataInclusionMask
        self.pix2uM = p2SRatio
        self.flowVoids = flowVoids * dataInclusionMask
        self.fvlabels, self.fvnlabels = ndimage.label(self.flowVoids)

    #===================================================================================================================
    #===================================================================================================================

    def vessRadiusStats(self):
        """
        Purpose:
            -Determine vessel radius at all locations along the vessel skeletionization
        Parameters:
            None
        Returns:
            -list of values of vessel radius at all locations along vessel skeletionization
        """

        minDistImg = ndimage.distance_transform_edt(self.vessels)
        minDistLs = [minDistImg[y,x] for (y,x) in np.ndindex(minDistImg.shape) if self.numberedSkel[y,x]!=0]
        minDistLs = [1 if m==0 else m for m in minDistLs]
        return np.array(minDistLs)*self.pix2uM

    #===================================================================================================================
    #===================================================================================================================

    def vessRadiusImage(self):
        """
        Purpose:
            -Visualize vessel radius at all locations along the vessel skeletionization
        Parameters:
            None
        Returns:
            -numpy array image of vessel skeleton with pixels coded by radius
        """

        minDistImg = ndimage.distance_transform_edt(self.vessels)
        for (y,x) in np.ndindex(minDistImg.shape):
            if self.numberedSkel[y,x]==0: minDistImg[y,x]=0
            elif minDistImg[y,x]==0: minDistImg[y,x]=1
        return minDistImg*self.pix2uM

    #===================================================================================================================
    #===================================================================================================================

    def branchNumberStats(self):
        """
        Purpose:
            -Determine branch number of every branch-point in image
        Parameters:
            None
        Returns:
            -list of branch numbers for each branch-point in image
        """

        branchNum = [len(n) for n in self.nodeGraph.values()]
        return branchNum

    #===================================================================================================================
    #===================================================================================================================

    def branchNumberImage(self):
        """
        Purpose:
            -Visualize branch number of every branch-point in image
        Parameters:
            None
        Returns:
            -numpy array image of vessel skeleton with pixels coded by branch number of closest branch-point
        """

        bnImg = np.zeros(self.numberedSkel.shape)
        #Searching through all connected branch-point branch-point pairs
        for n1 in self.nodeGraph.keys():
            for n2 in self.nodeGraph[n1]:
                #For each branch between the two nodes, find the halfway point and each half gets coded
                #   to the branch-number of the branch-point it is closer to
                if n2 not in self.nodeGraph.keys(): continue
                bnImg[n1[0],n1[1]]=len(self.nodeGraph[n1])
                bnImg[n2[0],n2[1]]=len(self.nodeGraph[n2])
                if (n1,n2) in self.branchGraph.keys():
                    bL = len(self.branchGraph[(n1,n2)])
                    branch = self.branchGraph[(n1,n2)]
                    for i in range(0,bL):
                        y,x = branch[i]
                        if i*2<bL:
                            bnImg[y,x]=len(self.nodeGraph[n1])
                        else:
                            bnImg[y,x]=len(self.nodeGraph[n2])
                else:
                    bL = len(self.branchGraph[(n2,n1)])
                    branch = self.branchGraph[(n2,n1)]
                    for i in range(0,bL):
                        y,x = branch[i]
                        if i*2<bL:
                            bnImg[y,x]=len(self.nodeGraph[n2])
                        else:
                            bnImg[y,x]=len(self.nodeGraph[n1])
        return bnImg

    #===================================================================================================================
    #===================================================================================================================

    def branchLengthStats(self):
        """
        Purpose:
            -Determine branch length of every vessel branch in image
        Parameters:
            None
        Returns:
            -list of branch lengths of all branches in image
        """

        branchLen = [len(b) for b in self.branchGraph.values()]
        return np.array(branchLen)*self.pix2uM

    #===================================================================================================================
    #===================================================================================================================

    def branchLengthImage(self):
        """
        Purpose:
            -Visualize branch length of every vessel branch in image
        Parameters:
            None
        Returns:
            -numpy array image of vessel skeleton with pixels coded by branch length of branch pixel is on
        """

        blImg = np.zeros(self.numberedSkel.shape)
        for branch in self.branchGraph.values():
            for b in branch:
                blImg[b[0],b[1]]=len(branch)
        for node in self.nodeGraph.keys():
            blImg[node[0],node[1]] = np.max(blImg[np.maximum(node[0]-1,0):np.minimum(node[0]+2,self.numberedSkel.shape[0]),np.maximum(node[1]-1,0):np.minimum(node[1]+2,self.numberedSkel.shape[1])])
        return blImg*self.pix2uM

    #===================================================================================================================
    #===================================================================================================================

    def flowVoidDistanceStats(self):
        """
        Purpose:
            -Determine flow void radius for every flow void in image
        Parameters:
            None
        Returns:
            -list of flow void radii of all flow voids in image
        """

        distList=[]
        minDistImg = ndimage.distance_transform_edt(self.flowVoids)
        for i in range(1,self.fvnlabels+1):
            spotMask = (self.fvlabels==i)*1
            minDistSelect = spotMask * minDistImg
            distList.append(np.max(minDistSelect))
        return np.array(distList)*self.pix2uM

    #===================================================================================================================
    #===================================================================================================================

    def flowVoidDistanceStats2(self):
        """
        Purpose:
            -Determine distance to nearest vessel at every location in flow voids
        Parameters:
            None
        Returns:
            -list of distance-to-nearest-vessel of every location in flow voids
        """

        distList=[]
        minDistImg = ndimage.distance_transform_edt(self.flowVoids)
        for i in range(1,self.fvnlabels+1):
            spotMask = (self.fvlabels==i)*1
            minDistSelect = spotMask * minDistImg
            distList = distList+[md for md in np.array(minDistSelect).flatten().tolist() if md!=0]
        return np.array(distList)*self.pix2uM

    #===================================================================================================================
    #===================================================================================================================

    def flowVoidDistanceImage(self):
        """
        Purpose:
            -Visualize distance to nearest vessel at every location in flow voids
        Parameters:
            None
        Returns:
            -numpy array image of flow voids coded by distance to the nearest vessel
        """

        return ndimage.distance_transform_edt(self.flowVoids)*self.pix2uM

    #===================================================================================================================
    #===================================================================================================================

    def flowVoidAreaStats(self):
        """
        Purpose:
            -Determine area of flow voids in image
        Parameters:
            None
        Returns:
            -list of area of all flow voids in image
        """

        labFlat = [fv for fv in self.fvlabels.flatten() if fv!=0]
        unique, areas = np.unique(labFlat, return_counts=True)
        return areas*(self.pix2uM*self.pix2uM)

    #===================================================================================================================
    #===================================================================================================================

    def flowVoidAreaImage(self,log=False):
        """
        Purpose:
            -Visualize  area of flow voids in image
        Parameters:
            None
        Returns:
            -numpy array image of flow voids coded by their area
        """

        fvaImg = np.zeros(self.numberedSkel.shape)
        fvList = self.flowVoidAreaStats()
        for y,x in np.ndindex(fvaImg.shape):
            fvaImg[y,x] = (self.fvlabels[y,x]!=0)*fvList[self.fvlabels[y,x]-1]
        if log:
            fvaImg = np.log(fvaImg+1)
        return fvaImg

#=======================================================================================================================
#\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
#=======================================================================================================================

class Point:
    """Data-holding class representing individual points in vessel skeleton. Each point knows what it's neighboring points are"""

    #===================================================================================================================
    #===================================================================================================================

    def __init__(self,coord):
        """
        Purpose:
            -Constructor method for generating Points.
        Parameters:
            -coords: (y,x) tuple of point coordinate (tuple)
        Returns:
            -None
        """

        self.y=coord[0]
        self.x=coord[1]
        self.neighbors=[]

    #===================================================================================================================
    #===================================================================================================================

    def __str__(self):
        """
        Purpose:
            -Print string method to dispaly point information
        Parameters:
            -None
        Returns:
            -String describing point
        """

        if len(self.neighbors)==0: statusS='Loner'
        elif len(self.neighbors)==1: statusS='EndPt'
        elif len(self.neighbors)==2: statusS='Branch'
        else: statusS='Node'
        return 'Point(%d,%d) || Status:%s | BranchN:%d ||'%(self.y,self.x,statusS,len(self.neighbors))

#=======================================================================================================================
#\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
#=======================================================================================================================

#Start of code execustion

#Create instances of classes for image processing
flHandler=filehandler()
imProcessor=imageProcessor()
skProcessor=skelProcessor()

#Read in config file with processing parameters
cD=flHandler.readConfig()

#Read in image file
fn = cD["picFileName"]
picBeforeProcessing=flHandler.readImage(fn,cD["picYRange"],cD["picXRange"])

#Pre-LMNM processing to blur image and remove low-intensity noise
picProcessing1=picBeforeProcessing
if cD["noiseSupressionTF"]:
    picProcessing1 = imProcessor.backgroundReduction(picProcessing1,cD["gaussLoop"],cD["gaussSize"],cD["exclusionMin"])

#Core methods of LMNM processing
xPeakProcessed,yPeakProcessed = imProcessor.processHorVerCombine(picProcessing1,cD["localPeakNormalizationType"],cD["peakSpacingMin"],cD["peakHeighDifferenceMin"],cD["filteredRatioCutoff"],cD["ratioCutoffLow"],cD["ratioCutoffHigh"])
combinedXYProcessed = imProcessor.xyCombine(xPeakProcessed,yPeakProcessed,cD["gaussLoop2"],cD["gaussSize2"])
imMask = imProcessor.mask(combinedXYProcessed,cD["maskingThresh"],cD["maskSizeExclusion"])

#Skeletonization of vessel mask
skeleton = imProcessor.skeletonize(imMask, cD["skeletonType"])

#Repairing incomplete and excessive skeletonization
skelNumbered = skProcessor.neighborCountMapper(skeleton)
if cD["skeletonBranchConnectingTF"]:
    skelNumbered = skProcessor.shortBranchConnector(skelNumbered,cD["sbcFitLength"],cD["sbcFitAngle"],cD["sbcFitExtension"])
if cD["skeletonBranchPruningTF"]:
    skelNumbered = skProcessor.shortBranchPruner(skelNumbered,cD["sbpMinAllowedLength"])
skeleton = (skelNumbered!=0)

#Convert skeleton into graph data-type representation
skelGraph = skProcessor.makeSkeletonGraph(skelNumbered) # (Y,X) tuple as keys for point
nodeGraph,branchGraph = skProcessor.makeNodeGraph(skelNumbered,skelGraph) # (Y,X) tuple as key for list of adjacent nodes, #((Y,X),(Y,X)) tuple of tuples as key for list of points in branch

#Optional masking to process only part of image
dataInclusionMask = np.full(skelNumbered.shape,1)
if cD["statMaskUse"]: dataInclusionMask = (flHandler.readImage(cD["statMaskFileName"],cD["picYRange"],cD["picXRange"])!=0)*1
if cD["invertMask"]: dataInclusionMask = (dataInclusionMask==0)*1

#Create instance of data analysis class with data generated via LMNM processing
skDataAnal=skelAnalysis(skelNumbered,(combinedXYProcessed>=cD["vessCutoff"])*1,((skelNumbered==0)*(combinedXYProcessed<cD["flowVoidCutoff"])*1),skelGraph,nodeGraph,branchGraph,dataInclusionMask,cD["pixelTouMRatio"])

#Image analysis
vRad = skDataAnal.vessRadiusStats()
bNum = skDataAnal.branchNumberStats()
bLen = skDataAnal.branchLengthStats()
fDis = skDataAnal.flowVoidDistanceStats()
fDis2 =skDataAnal.flowVoidDistanceStats2()
fAre = skDataAnal.flowVoidAreaStats()
print '   vessel radius: \t',np.mean(vRad),'\t',np.median(vRad),'\t',np.std(vRad)*2/np.sqrt(len(vRad))
print '   branch number: \t',np.mean(bNum),'\t',np.median(bNum),'\t',np.std(bNum)*2/np.sqrt(len(bNum))
print '   branch length: \t',np.mean(bLen),'\t',np.median(bLen),'\t',np.std(bLen)*2/np.sqrt(len(bLen))
print ' flowvoid radius: \t',np.mean(fDis),'\t',np.median(fDis),'\t',np.std(fDis)*2/np.sqrt(len(fDis))
print 'flowvoid radius2: \t',np.mean(fDis2),'\t',np.median(fDis2),'\t',np.std(fDis2)*2/np.sqrt(len(fDis2))
print '   flowvoid area: \t',np.mean(fAre),'\t',np.median(fAre),'\t',np.std(fAre)*2/np.sqrt(len(fAre)),'\n'

#Generation of images to describe image analysis
vRadI = skDataAnal.vessRadiusImage()
bNumI = skDataAnal.branchNumberImage()
bLenI = skDataAnal.branchLengthImage()
fDisI = skDataAnal.flowVoidDistanceImage()
fAreI = skDataAnal.flowVoidAreaImage()

#Saves skeleton overlaid on origional image of vessels if saveImg is turned to true
saveImg = False
if saveImg:
    plt.axis('off')
    plt.imshow(picBeforeProcessing,cmap='gray')
    ls=[]
    for y,x in np.ndindex(skeleton.shape):
        if skeleton[y,x]:
            ls.append((y,x))
    plt.plot([l[1] for l in ls],[l[0] for l in ls],'ro',ms=2)
    fig = plt.gcf()
    fig.set_size_inches(10,10)
    fig.savefig('%s_skel.png'%(fn[:-4]), dpi=400)
    plt.clf()

#Displays origional image as well as visual representations of data analysis
a=plt.subplot2grid((2, 3), (0, 0))
plt.imshow(picBeforeProcessing,cmap='gray')
ls=[]
for y,x in np.ndindex(skeleton.shape):
    if skeleton[y,x]:
        ls.append((y,x))
plt.plot([l[1] for l in ls],[l[0] for l in ls],'ro',ms=2)
plt.subplot2grid((2, 3), (0, 1),sharex=a,sharey=a)
plt.imshow(vRadI,cmap='jet')
plt.subplot2grid((2, 3), (0, 2),sharex=a,sharey=a)
plt.imshow(bNumI,cmap='jet')
plt.subplot2grid((2, 3), (1, 0),sharex=a,sharey=a)
plt.imshow(bLenI,cmap='jet')
plt.subplot2grid((2, 3), (1, 1),sharex=a,sharey=a)
plt.imshow(fDisI,cmap='jet')
plt.subplot2grid((2, 3), (1, 2),sharex=a,sharey=a)
plt.imshow(fAreI,cmap='jet')

plt.show()
